#!/usr/bin/env python3

## standard library
from os import getenv
## third party
from botocore.exceptions import NoCredentialsError
from rich.console import Console
from rich.table import Table
import boto3
## custom
from modules.Auth import GkeAwsSession
from modules.Helpers import CollectionTarget, loadYaml

console = Console()


def print_secrets(client: 'boto3.client.secretsmanager', namePrefix='mcs2/', max=5):
    paginator = client.get_paginator('list_secrets')
    response_iterator = paginator.paginate(
        Filters=[
            {
                'Key': 'name',
                'Values': [namePrefix]
            },
        ],
        SortOrder='asc',
        PaginationConfig = { 'MaxItems': max } if max else {}
    )
    columns = ['AWS Secret Name', 'ARN']
    table = Table(title=f'AWS Secrets ({client.meta.region_name})', show_lines=True, title_style='bold magenta', *columns)
    for page in response_iterator:
        for i in page["SecretList"]:
            table.add_row(i['Name'], i['ARN'])
    console.print(table)


def main(
        audience,
        role_arn,
        config_file,
        ):
    try:
        client = boto3.client('sts')
        client.get_caller_identity()
    except NoCredentialsError:
        session = GkeAwsSession(
            audience = audience,
            role_arn = role_arn,
            region = getenv('AWS_DEFAULT_REGION', 'us-west-2')
            )
        boto3.setup_default_session(botocore_session=session)

    ## whoami
    client = boto3.client('sts')
    client.get_caller_identity()
    ## 'set_config_variable' above sets the default region: us-west-2
    client.meta.region_name

    ## get Secrets from in us-west-2
    client_usw2 = boto3.client('secretsmanager')
    client_usw2.meta.region_name
    print_secrets(client_usw2)

    ## get Secrets from in us-east-1
    client_use1 = boto3.client('secretsmanager', region_name='us-east-1')
    client_use1.meta.region_name
    print_secrets(client_use1)

    ## loop the config; this this is a CM mounted as a file inside the container
    config = loadYaml(config_file)
    try:
        collection = config["collection"]
    except KeyError:
        raise
    for k,v in collection.items():
        target = CollectionTarget(name=k, **v)
        if target.aws_replication:
            console.print(f'Replicating to AWS: {target}', style='light_salmon1')
        else:
            console.print(f'Not replicating to AWS: {target}', style='light_salmon1')


if __name__ == '__main__':
    main(
        audience = __file__,
        role_arn = 'arn:aws:iam::605728677638:role/test-secrets-sync',
        config_file = '/opt/config.yaml',
    )