#!/usr/bin/env python3

"""
Script to copy AWS Secrets Manager secrets to GCP Secrets Manager using the newest naming convention.
"""

from contextlib import redirect_stdout
from typing import List
from typing_extensions import Annotated
import io
import logging
import re

## third-party
from google.cloud import secretmanager
from parse import parse
from rich.console import Console
from rich.table import Table
import boto3
import google.auth
import typer
import warnings
warnings.filterwarnings("ignore", "Your application has authenticated using end user credentials")

## custom modules
from modules.Auth import checkAwsAuth
from modules.Helpers import AwsCollectionTarget, GcpCollectionTarget, loadYaml, Platform
from modules.Logger import *
from modules.ManagedSecrets import Mcs2AwsSecret, GcpSecret
from modules.Normalization import awsSecretsParser


## Gobals
logger = logging.getLogger(__name__)
logger.setLevel(level=logging.DEBUG)
app = typer.Typer(pretty_exceptions_show_locals=False)
console = Console()


def _getAwsSecrets(
        namePrefix: str,
        max: int = None,
        region: str = 'us-west-2',
        ) -> List[str]:
    client = boto3.client('secretsmanager', region_name = region)
    paginator = client.get_paginator('list_secrets')
    response_iterator = paginator.paginate(
        Filters=[
            {
                'Key': 'name',
                'Values': [namePrefix]
            },
        ],
        SortOrder='asc',
        PaginationConfig = { 'MaxItems': max } if max else {}
    )
    results = []
    for page in response_iterator:
        for i in page["SecretList"]:
            name = i['Name']
            env, productTeam, cmSubService, uniqueSecret = awsSecretsParser(name)
            ## avoid secrets with these patterns (for now)
            pattern = re.compile('(?i)(?:wavefront|hbfw|Sfdc|ivani4|testsecret|kenna|sensor-key-pair|ml-ops|CloudfrontKeyPair|atlas-configuration)|(/token$)|(\\btest\\b)', re.IGNORECASE)
            if pattern.search(name):
                console.print(f'skipping {name} (reason: regex pattern match)', style='light_salmon1')
                continue
            else:
                try:
                    ## skip instances that originate in GCP -- using presence of 'gcp-origin-name' tag to filter
                    next(item for item in i['Tags'] if item['Key'] == 'gcp-origin-name')['Value']
                    console.print(f'skipping {name} (reason: gcp-origin-name tag found)', style='light_salmon1')
                    continue
                except Exception:
                    console.print(f'processing {name}', style='slate_blue1')
            secretValue = client.get_secret_value(SecretId=name)['SecretString']
            asec = Mcs2AwsSecret(
                    env = env,
                    productTeam = productTeam,
                    cmSubService = cmSubService,
                    uniqueSecret = uniqueSecret,
                    normalized = False,
                )
            ## we set the `value` property here because the new, normalized AWS secret name doesn't exist yet
            asec.value = secretValue
            ## making up a custom property so I can use it later in this specific script's workflow
            asec.originalName = name
            results.append(asec)
    return results


def _getGcpSecrets(
        namePrefix: str,
        ) -> List[str]:
    _, project = google.auth.default()
    parent = f"projects/{project}"
    matches = []
    pattern = re.compile('projects/\\d*/secrets/%s' % namePrefix, re.IGNORECASE)
    logger.debug(f'matching pattern: {pattern}')
    client = secretmanager.SecretManagerServiceClient()
    for secret in client.list_secrets(request={"parent": parent, "filter": 'NOT labels.aws-origin-name:*'}):
        if pattern.match(secret.name):
            _, name = parse('projects/{}/secrets/{}', secret.name)
            logger.debug(f"Adding secret to the match list: {name}")
            matches.append(name)
    return matches


@app.command()
def CopyAwsSecrets(
        prefix: str,
        update: Annotated[bool, typer.Option(envvar="UPDATE", case_sensitive=False)] = False,
        inspect: Annotated[bool, typer.Option(envvar="INSPECT", case_sensitive=False)] = False,
        region: str = 'us-west-2'
        ) -> None:
    secrets = _getAwsSecrets(prefix, region=region)
    columns = ['AWS Secret Original Name', 'AWS Secret Name', 'GCP Secret', 'Status Message(s)']
    table = Table(title='Copy AWS Secrets', show_lines=True, title_style='bold magenta', *columns)
    for asec in secrets:
        if inspect:
            console.print(asec.value, style='bright_yellow') if asec.value else console.print('Class value is not set.', style='bright_yellow')
        gsec = GcpSecret(
            env = asec.env.name,
            productTeam = asec.productTeam,
            cmSubService = asec.cmSubService,
            uniqueSecret = asec.uniqueSecret,
        )
        gsec.value = asec.value
        if inspect:
            console.print(gsec.value, style='bright_yellow') if gsec.value else console.print('Class value is not set.', style='bright_yellow')
            console.print(gsec.secretValue, style='slate_blue1') if gsec.secretValue else console.print('Secret value is not set.', style='slate_blue1')
        f = io.StringIO()
        with redirect_stdout(f):
            if update:
                labels = {
                    'aws-origin-name': re.sub("/", '-', asec.originalName.lower()),
                    'aws-origin-region':  asec.region
                }
                gsec.update(labels=labels)
            else:
                gsec.compare()
        out = f.getvalue()
        table.add_row(asec.originalName, asec.name, gsec.name, out)
    console.print(table)
    print(len(secrets))


@app.command()
def CopyGcpSecrets(
        prefix: str,
        update: Annotated[bool, typer.Option(envvar="UPDATE", case_sensitive=False)] = False,
        inspect: Annotated[bool, typer.Option(envvar="INSPECT", case_sensitive=False)] = False,
        region: str = 'us-west-2'
        ) -> None:
    logger.debug('Command Client region: %s' % boto3.client('sts').meta.region_name)
    secrets = _getGcpSecrets(namePrefix=prefix)
    columns = ['GCP Secret', 'AWS Secret', 'Status Message(s)']
    table = Table(title='Copy GCP Secrets', show_lines=True, title_style='bold magenta', *columns)
    for name in secrets:
        env, productTeam, cmSubService, uniqueSecret = parse('{}-{}-{}-{}', name)
        gsec = GcpSecret(
            env = env,
            productTeam = productTeam,
            cmSubService = cmSubService,
            uniqueSecret = uniqueSecret,
        )
        if not gsec.secret:
            raise Exception('Specified GCP Secret does not exist!')
        else:
            if inspect:
                console.print(gsec, style='slate_blue1')
                console.print(gsec.secretValue, style='slate_blue1') if gsec.secretValue else console.print('Secret value is not set.', style='slate_blue1')
                console.print(gsec.value, style='bright_yellow') if gsec.value else console.print('Class value is not set.', style='bright_yellow')
                console.print("-----------------------", style='bright_magenta')
            asec = Mcs2AwsSecret(
                env = env,
                productTeam = productTeam,
                cmSubService = cmSubService,
                uniqueSecret = uniqueSecret,
                normalized = False,
            )
            class_region = asec.smClient.meta.region_name
            logger.debug('Class Client region: %s' % class_region)
            if class_region != region:
                asec.region = region
                logger.debug('New Class Client region: %s' % asec.smClient.meta.region_name)
            logger.debug(asec.arn)
            asec.value = gsec.value
            if inspect:
                console.print(asec, style='slate_blue1')
                console.print(asec.value, style='bright_yellow') if asec.value else console.print('Class value is not set.', style='bright_yellow')
                console.print(asec.secretValue, style='slate_blue1') if asec.secretValue else console.print('Secret value is not set.', style='slate_blue1')
            f = io.StringIO()
            with redirect_stdout(f):
                if update:
                    tags = {
                        'gcp-origin-name': gsec.name
                    }
                    awsTags = [{ 'Key': k, 'Value': v } for k,v in tags.items()]
                    asec.update(addtlTags=awsTags)
                else:
                    asec.compare()
            out = f.getvalue()
            table.add_row(gsec.name, asec.name, out)
    console.print(table)


@app.command()
def CopySecretsViaHelm(
        platform: Annotated[ Platform, typer.Argument(case_sensitive=False)],
        config_file: str = 'test_config.yaml',
        update: Annotated[bool, typer.Option(envvar="UPDATE", case_sensitive=False)] = False,
        inspect: Annotated[bool, typer.Option(envvar="INSPECT", case_sensitive=False)] = False,
        ) -> None:
    ## loop the config; this this is a CM mounted as a file inside the container
    config = loadYaml(config_file)
    try:
        collection = config["collection"]
    except KeyError:
        raise
    if platform == 'gcp':
        for k,v in collection.items():
            target = GcpCollectionTarget(name=k, **v)
            if target.aws_replication:
                CopyGcpSecrets(target.name, region=target.region, update=update, inspect=inspect)
            else:
                console.print(f'Not replicating to AWS: {target}', style='light_salmon1')
    elif platform == 'aws':
        for k,v in collection.items():
            target = AwsCollectionTarget(name=k, **v)
            CopyAwsSecrets(target.name, region=target.region, update=update, inspect=inspect)
    else:
        raise Exception('Unmanaged Platform Selected')


if __name__ == "__main__":
    checkAwsAuth()
    app()
