#!/usr/bin/env python3

## standard library
from os import getenv
import socket
## third-party modules
from botocore.credentials import RefreshableCredentials
from botocore.exceptions import NoCredentialsError
from botocore.session import get_session, Session
from google.auth.transport.requests import Request as GoogleAuthRequest
import boto3
import google.auth
import google.oauth2.id_token
import requests
## custom modules
from modules.Logger import *


logger = logging.getLogger(__name__)
logger.setLevel(level=logging.DEBUG)

# GKE Token Fetcher

def checkAwsAuth(region: str = 'us-west-2'):
    try:
        client = boto3.client('sts')
        _ = client.get_caller_identity()
        logger.debug('Existing credential detected')
        if client.meta.region_name != region:
            logger.debug(f'Changing region to {region}')
            boto3.setup_default_session(region_name=region)
    except NoCredentialsError:
        audience = getenv('OPENID_AUDIENCE', 'sm-util')
        role_arn = getenv('AWS_ROLE_ARN', None)
        if isGKE() and role_arn:
            logger.debug(f'Creating a new GKE session for region: {region}')
            session = GkeAwsSession(
                audience = audience,
                role_arn = role_arn,
                region = region
                )
            boto3.setup_default_session(botocore_session=session)
        else:
            raise Exception('No AWS Credentials detected.')


def whoami() -> None:
    client = boto3.client('sts')
    return client.get_caller_identity()


def isGKE():
    try:
        _ = socket.getaddrinfo("metadata.google.internal", 80, proto=socket.IPPROTO_TCP)
        return True
    except socket.gaierror:
        return False


def fetch_svc_acount_email(service_account : str = 'default') -> str:
    """
    Retrieves Default Service Account email from the metadata server.

    Returns:
        Google Service Account email address
    """
    metadata_url = "http://metadata.google.internal/computeMetadata/v1/"
    headers = {'Metadata-Flavor': 'Google'}
    url = f"{metadata_url}/instance/service-accounts/{service_account}/email"
    r = requests.get(url, headers=headers)
    r.raise_for_status()
    r.content


def fetch_svc_acount_access_token(service_account : str = 'default') -> str:
    """
    Retrieves access token from the metadata server.

    Returns:
        The access token.
    """
    metadata_url = "http://metadata.google.internal/computeMetadata/v1"
    headers = {'Metadata-Flavor': 'Google'}
    url = f"{metadata_url}/instance/service-accounts/{service_account}/token"
    # Request an access token from the metadata server.
    r = requests.get(url, headers=headers)
    r.raise_for_status()
    # Extract the access token from the response.
    access_token = r.json()["access_token"]
    return access_token


def fetch_id_token(target_audience : str) -> str:
    """
    Fetch the ID Token / JWT (Json Web Token) from the current environment.

    See https://googleapis.dev/python/google-auth/1.14.0/reference/google.oauth2.id_token.html#google.oauth2.id_token.fetch_id_token

    audience (str) – The audience that this ID token is intended for. This is part of the AWS Role's Trusted Identity.

    Returns:
        The an OpenID Connect ID Tokens.
    """
    id_token = google.oauth2.id_token.fetch_id_token(GoogleAuthRequest(), target_audience)
    return id_token


def fetch_id_token_details(id_token : str) -> None:
    """
    Verifies and Fetches the ID Token details

    Returns:
        The an OpenID Connect ID Tokens.
    """
    resp = google.oauth2.id_token.verify_token(request=GoogleAuthRequest(), id_token=id_token)
    print(resp)


def fetch_gke_token(service_account : str, audience : str) -> str:
    """
    DEPRECATED

    Fetch the ID Token from the current environment given a specific service account and target audience string.

    Limitations here are whether the current context can access the service account.

    Audience: The audience that this ID token is intended for. It must be one of the OAuth 2.0 client
    IDs of your application.

    Returns:
        The an OpenID Connect ID Tokens.
    """
    url = f"http://metadata/computeMetadata/v1/instance/service-accounts/{service_account}/identity?audience={audience}"
    headers = {'Metadata-Flavor': 'Google'}
    r = requests.get(url, headers=headers)
    if r.status_code != 200:
        raise Exception(f"unexpected status code: {r.status_code}")
    return r.content.decode('utf-8')


# Refreshable Credentials Provider
def _assume_role_with_web_identity(gke_token : bytes,
                                   role_arn : str,
                                   RoleSessionName='GKEWebIdentitySession',
                                   duration=3600):
    sts_client = boto3.client('sts')
    response = sts_client.assume_role_with_web_identity(
        RoleArn=role_arn,
        RoleSessionName=RoleSessionName,
        WebIdentityToken=gke_token,
        DurationSeconds=duration
    )
    credentials = response['Credentials']
    return {
        'access_key': credentials['AccessKeyId'],
        'secret_key': credentials['SecretAccessKey'],
        'token': credentials['SessionToken'],
        'expiry_time': credentials['Expiration'].isoformat()
    }


def GkeAwsSession(audience : str, role_arn : str, region : str) -> Session:
    """
    Retrieves a JWT token from GKE environment and turns it into an AWS Session
    that can be used to form AWS Service Clients.

    Returns:
        botocore.session.Session
    """
    gke_token = fetch_id_token(audience)
    session = get_session()
    refreshable_credentials = RefreshableCredentials.create_from_metadata(
        metadata=_assume_role_with_web_identity(gke_token, role_arn),
        refresh_using=_assume_role_with_web_identity,
        method='sts-assume-role-with-web-identity'
    )
    session._credentials = refreshable_credentials
    session.set_config_variable('region', region)
    return session
