#!/usr/bin/env python3

from dataclasses import dataclass
from typing import Literal, List
import re

## third-party
from botocore.session import Session
from google.cloud import resourcemanager_v3 # google-cloud-resource-manager
from google.cloud import secretmanager_v1 # google-cloud-secret-manager
from google.cloud.secretmanager_v1.types import Secret, SecretVersion, SecretPayload
import boto3
import google.api_core.exceptions
import google.auth
import warnings
warnings.filterwarnings("ignore", "Your application has authenticated using end user credentials")

## shared custom modules
from modules.Helpers import parseString
from modules.Logger import *


logger = logging.getLogger(__name__)
logger.setLevel(level=logging.DEBUG)


@dataclass
class CbcEnvironment:
    _name: str

    def __str__(cls):
        return f'{type(cls).__name__}({cls.name})'

    def __post_init__(cls):
        cls._name = cls._name.lower()

    @property
    def name(cls):
        try:
            return {
                "kcdev01":   "dev01",
                "kctest03":  "test03",
                "eap1use1":  "eap01",
                "prd1use1":  "prod01",
                "prd2use1":  "prod02",
                "prd5use1":  "prod05",
                "prd1ec1":   "prod06",
                "prd1apne1": "prodnrt",
                "prd1apse2": "prodsyd",
                "prd1apse2": "prodsyd",
            }[cls._name]
        except KeyError:
            return cls._name

    @property
    def kcenv(cls):
        try:
            return {
                "dev01": "kcdev01",
                "test03": "kctest03",
                "eap01": "eap1use1",
                "prod01": "prd1use1",
                "prod02": "prd2use1",
                "prod05": "prd5use1",
                "prod06": "prd1ec1",
                "prodnrt": "prd1apne1",
                "prodsyd": "prd1apse2",
                "prodsyd": "prd1apse2",
            }[cls.name]
        except KeyError:
            return cls._name



@dataclass
class Mcs2AwsSecret:
    env: str
    productTeam: str
    cmSubService: str
    uniqueSecret: str
    normalized: bool = False
    normalize_option: Literal['title', 'snake'] = 'title'
    _value = None
    _secret = None
    _awsSession = None
    _region = None
    _smClient = None

    def __str__(cls):
        s = ', '.join(f'{field}={getattr(cls, field)}' for field in ['env', 'productTeam', 'cmSubService', 'uniqueSecret'])
        return f'{type(cls).__name__}({s})'

    def __post_init__(cls):
        cls.env = CbcEnvironment(cls.env)
        if cls.normalized:
            object.__setattr__(cls, 'uniqueSecret', cls.normalize(parseString(cls.uniqueSecret)))

    @property
    def name(cls):
        """ The AWS Secret Name. """
        return f'mcs2/{cls.env.kcenv}/{cls.id}'

    @property
    def id(cls):
        """ The string used to identify the secret within an environment scope. """
        return '-'.join(item for item in [cls.productTeam, cls.cmSubService, cls.uniqueSecret] if (item and item != ''))

    @property
    def value(cls):
        if not cls._value:
            return cls.secretValue
        else:
            return cls._value

    @value.setter
    def value(cls, v) -> None:
        """ Sets the class value explicitly. """
        cls._value = v

    @property
    def secret(cls) -> dict:
        if not cls._secret:
            return cls._get()
        else:
            return cls._secret

    @secret.setter
    def secret(cls, v: dict) -> None:
        cls._secret = v

    @property
    def region(cls) -> str:
        """ The AWS Region """
        if not cls._region:
            return boto3.client('sts').meta.region_name
        else:
            return cls._region

    @region.setter
    def region(cls, v : str = None) -> None:
        """ Sets the AWS Region """
        #TODO - validate againsts known AWS regions
        cls._region = v

    @property
    def smClient(cls):
        """ An AWS Client for SecretsManager """
        args={}
        if cls._region:  args["region_name"] = cls._region
        return boto3.client('secretsmanager', **(args))

    @property
    def awsSession(cls) -> Session:
        """ An AWS Botocore Session """
        args={}
        if cls._region:  args["region_name"] = cls._region
        logger.debug("AWS Session Args: %s" % args)
        return boto3.session.Session(**(args))

    @property
    def arn(cls):
        """ ARN for the Secret """
        try:
            return cls.secret['ARN']
        except TypeError:
            return f'ARN: Secret {cls.name} not found in region ({cls.region})'

    def _get(cls) -> dict:
        args={ 'Filters' : [{'Key': 'name', 'Values': [cls.name]}] }
        response = cls.smClient.list_secrets(**(args))
        secretList = response['SecretList']
        if len(secretList) > 1:
            raise Exception('Too many secrets returned.')
        elif len(secretList) == 0:
            return None
        else:
            return secretList[0]

    def _create(cls, addtlTags: List[dict] = None) -> None:
        """ Creates the Secret container in AWS if it doesn't exist. """
        tags = [
            {
                'Key': "env",
                'Value': cls.env.name
            },
            {
                'Key': "productTeam",
                'Value': cls.productTeam
            },
            {
                'Key': "cmSubService",
                'Value': cls.cmSubService
            },
            {
                'Key': "uniqueSecret",
                'Value': cls.uniqueSecret
            },
        ]
        if addtlTags:
            tags.extend(addtlTags)
        response = cls.smClient.create_secret(
            Name=cls.name,
            Description=f'{cls.uniqueSecret} for this mcs2 environment',
            SecretString=cls.value,
            Tags=tags,
        )
        return response

    @property
    def secretValue(cls, VersionId: str = None, VersionStage: str = None):
        """ Retrieves the value of the latest Secret Versions from GCP if it exists. """
        args={ 'SecretId' : cls.name }
        if VersionId: args['VersionId'] = VersionId
        if VersionStage: args['VersionStage'] = VersionStage
        try:
            response = cls.smClient.get_secret_value(**(args))
            return response['SecretString']
        except cls.smClient.exceptions.ResourceNotFoundException:
            ## there are no versions
            return None

    def compare(cls):
        """ Checks to see if the Class value matches the Secret Value. """
        if not cls.value:
            print('Class value is not set.')
            return False
        elif not cls.secretValue:
            print('Secret value is not set.')
            return False
        else:
            if cls.value != cls.secretValue:
                print("The existing class value is different than the latest AWS Secret Version's value")
                return False
            else:
                print(f'The latest Secret version value matches.')
                return True

    def update(cls, VersionStages: List[str] = None, addtlTags: List[dict] = None) -> None:
        """ Sets the Secret Value if it's different than the Class value. """
        if not cls.compare():
            print('Updating the Secret Value...')
            args={
                'SecretId' : cls.name,
                'SecretString' : cls.value,
            }
            if VersionStages: args['VersionStages'] = VersionStages
            try:
                cls.smClient.put_secret_value(**(args))
            except cls.smClient.exceptions.ResourceNotFoundException:
                cls._create(addtlTags=addtlTags)
                cls.smClient.put_secret_value(**(args))
        else:
            pass
        if addtlTags:
            cls.smClient.tag_resource(
                SecretId=cls.name,
                Tags=addtlTags
            )

    def normalize(cls, items: List, option = None):
        elements = []
        for i in items:
            if re.search('-|_', i) and isinstance(i, str):
                elements.extend(parseString(i))
            else:
                elements.append(i)
        if re.search('snake', option if option else cls.normalize_option, re.IGNORECASE):
            return '_'.join(e for e in elements)
        else:
            ## example: Title case
            return ''.join(e.title() for e in elements)



@dataclass
class GcpSecret:
    env: str
    productTeam: str
    cmSubService: str
    uniqueSecret: str
    _value = None
    _secret = None
    smClient = secretmanager_v1.SecretManagerServiceClient()
    projClient = resourcemanager_v3.ProjectsClient()
    ## TODO: method for comparing most recent update in UTC

    def __str__(cls):
        s = ', '.join(f'{field}={getattr(cls, field)}' for field in ['env', 'productTeam', 'cmSubService', 'uniqueSecret'])
        return f'{type(cls).__name__}({s})'

    def __post_init__(cls):
        cls.env = CbcEnvironment(cls.env)

    @property
    def projectId(cls) -> str:
        _, project = google.auth.default()
        return project

    @property
    def projectNumber(cls) -> int:
        request = resourcemanager_v3.SearchProjectsRequest(query=f"id:{cls.projectId}")
        page_result = cls.projClient.search_projects(request=request)
        for response in page_result:
            if response.project_id == cls.projectId:
                return int(response.name.replace('projects/', ''))
        # if not yet returned, raise
        raise Exception('Could not identify the correct project number')

    @property
    def id(cls):
        """ The string used to identify the secret within a parent project. """
        return '-'.join(item for item in [cls.env.name, cls.productTeam, cls.cmSubService, cls.uniqueSecret] if (item and item != ''))

    @property
    def name(cls):
        """ The GCP Secret Full Name. """
        return f'projects/{cls.projectId}/secrets/{cls.id}'

    @property
    def value(cls):
        if not cls._value:
            return cls.secretValue
        else:
            return cls._value

    @value.setter
    def value(cls, v) -> None:
        """ Sets the class value explicitly. """
        cls._value = v

    @property
    def payload(cls) -> SecretPayload:
        """ Transforms the Class value into a GCP Payload object. """
        ## might need to cast this into JSON
        value = cls.value.encode("UTF-8")
        return secretmanager_v1.SecretPayload(
            data=value
        )

    @property
    def secret(cls) -> Secret:
        if not cls._secret:
            return cls._get()
        else:
            return cls._secret

    @secret.setter
    def secret(cls, v: Secret) -> None:
        cls._secret = v

    def _get(cls) -> Secret:
        """ Retrieves the Secret container in GCP. """
        try:
            request = secretmanager_v1.GetSecretRequest(name=cls.name)
            secret = cls.smClient.get_secret(request=request)
            return secret
        except google.api_core.exceptions.NotFound:
            return None

    def _create(cls, labels: dict = None) -> None:
        """ Creates the Secret container in GCP if it doesn't exist. """
        ##TODO: normalize label keys and values to match [\p{Ll}\p{Lo}][\p{Ll}\p{Lo}\p{N}_-]{0,62}
        # defaults = {
        #     'env': cls.env.name,
        #     'product-team': cls.productTeam,
        #     'cm-sub-service': cls.cmSubService,
        #     'unique-secret': cls.uniqueSecret,
        # }
        # labels = defaults | labels
        secret = cls.smClient.create_secret(
            request={
                'parent': f'projects/{cls.projectId}',
                'secret_id': cls.id,
                'secret': {
                    'replication': { 'automatic': {} },
                    'labels': labels,
                },
            }
        )
        return secret

    def _updateLabels(cls, labels: dict = None) -> None:
        """ Adds labels to the existing GCP Secret """
        ##TODO: normalize label keys and values to match [\p{Ll}\p{Lo}][\p{Ll}\p{Lo}\p{N}_-]{0,62}
        labelDict = dict(cls.secret.labels) | labels
        secret = {"name": cls.secret.name, "labels": labelDict}
        update_mask = {"paths": ["labels"]}
        _ = cls.smClient.update_secret(
            request={"secret": secret, "update_mask": update_mask}
        )

    def _getVersion(cls, version) -> SecretVersion:
        request = secretmanager_v1.GetSecretVersionRequest( name=f'{cls.name}/versions/{version}' )
        return cls.smClient.get_secret_version(request=request)

    @property
    def latestVersion(cls) -> SecretVersion:
        """ Retrieves the latest Secret Version from GCP if it exists. """
        if cls.secret:
            request = secretmanager_v1.ListSecretVersionsRequest( parent=cls.secret.name )
            page_result = cls.smClient.list_secret_versions(request=request)
            if page_result.total_size != 0:
                sv = cls._getVersion('latest')
                ## find most-recent version that's enabled
                while sv.state != sv.state.ENABLED:
                    versionId = int(sv.name.split('/')[-1])
                    sv = cls._getVersion(versionId-1)
                return sv
            else:
                return None
        else:
            return None

    @property
    def secretValue(cls):
        """ Retrieves the value of the latest Secret Versions from GCP if it exists. """
        if cls.latestVersion:
            request = secretmanager_v1.AccessSecretVersionRequest( name=cls.latestVersion.name )
            versionResponse = cls.smClient.access_secret_version(request=request)
            return versionResponse.payload.data.decode("UTF-8")
        else:
            ## there are no versions
            return None

    def compare(cls):
        """ Checks to see if the Class value matches the Secret Value. """
        if not cls.value:
            print('Class value is not set.')
            return False
        elif not cls.secretValue:
            print('Secret value is not set.')
            return False
        else:
            if cls.value != cls.secretValue:
                print("The existing class value is different than the latest GCP Secret Version's value")
                return False
            else:
                print("The Secret Version value matches.")
                return True

    def update(cls, labels: dict = None):
        """ Sets the Secret Value if it's different than the Class value. """
        if not cls.compare():
            if not cls.secret:
                cls._create(labels=labels)
            print('Updating the Secret Value...')
            ## no larger than 64KiB
            request = secretmanager_v1.AddSecretVersionRequest(
                parent=cls.secret.name,
                payload=cls.payload,
            )
            return cls.smClient.add_secret_version(request=request)
        else:
            ##TODO: add a 'awsName' label that indicates it was copied
            if labels:
                cls._updateLabels(labels)
            return cls.latestVersion

