#!/usr/bin/env python3


from typing import Literal
import re
from xxlimited import foo
from parse import parse

## shared custom modules
from modules.Helpers import parseString
from modules.Logger import *


logger = logging.getLogger(__name__)
logger.setLevel(level=logging.DEBUG)


def normalizeCasing(item, option: Literal['title', 'snake'] = 'title'):
    logger.debug('elements to normalize: %s' % item)
    if isinstance(item, str):
        elements = parseString(item)
    elif isinstance(item, list):
        elements = []
        for i in item:
            if re.search('-|_', i) and isinstance(i, str):
                logger.debug('working with component: %s' % i)
                elements.extend(parseString(i))
            else:
                elements.append(i)
    else:
        raise Exception('Cannot normalize type: %s.' % type(item))
    if re.search('snake', option, re.IGNORECASE):
        return '_'.join(e for e in elements)
    else:
        ## example: Title case
        return ''.join(e.title() for e in elements)


def normalizeUniqueSecret(secretId: str, serviceName: str, exclusions: list[str] = None):
    elements = parseString(secretId)
    if exclusions:
        for item in exclusions:
            elements.remove(item)
    if '-' in serviceName:
        cmSubService = normalizeCasing(serviceName)
        ## have to search for each string b/c the separator is not included in the elements list
        remove_me_regex = '|'.join(e for e in parseString(cmSubService))
        [elements.remove(item) for item in parseString(secretId) if re.search(remove_me_regex, item, re.IGNORECASE)]
        uniqueSecret = normalizeCasing(elements)
    else:
        cmSubService = serviceName
        [elements.remove(item) for item in parseString(secretId) if re.search(cmSubService, item, re.IGNORECASE)]
        uniqueSecret = normalizeCasing(elements)
    return cmSubService, uniqueSecret


def awsSecretsParser(secretName: str):
    """
    Opinionated Parsing rules for existing AWS Secrets.
    ** Note the use of 're.match' (prefix matching) and 're.search' (whole string search)
    """
    env, uniqueSecret = parse('mcs2/{}/{}', secretName)
    ## case-by-case handling continues from here ...
    ## Starting with team-specific product matching
    if re.search('wavefront|logzio|artifactory|cb-vmware-com', uniqueSecret, re.IGNORECASE):
        ''' matching on known prodeng products '''
        productTeam = 'prodeng'
        pattern = 'wavefront|logzio|artifactory|cb-vmware-com'
        logger.debug('rule match: %s (%s)' % (pattern.upper(), uniqueSecret))
        match = re.search(pattern, uniqueSecret, re.IGNORECASE)
        cmSubService, uniqueSecret = normalizeUniqueSecret(uniqueSecret, match.group())
    elif re.search('mailsender', uniqueSecret, re.IGNORECASE):
        ''' matching on known Bolt products '''
        productTeam = 'bolt'
        pattern = 'mailsender'
        logger.debug('rule match: %s (%s)' % (pattern.upper(), uniqueSecret))
        match = re.search(pattern, uniqueSecret, re.IGNORECASE)
        cmSubService, uniqueSecret = normalizeUniqueSecret(uniqueSecret, match.group())
    elif re.search('appliance|vulnerability-assessment', uniqueSecret, re.IGNORECASE):
        ##TODO: candidate for release-name-based parsing (e.g. 'infra' is functionally a cm3 value which modifies 'vulnerability-assessment')
        ''' matching on known CWP products '''
        productTeam = 'cwp'
        pattern = 'appliance|vulnerability-assessment'
        logger.debug('rule match: %s (%s)' % (pattern.upper(), uniqueSecret))
        match = re.search(pattern, uniqueSecret, re.IGNORECASE)
        cmSubService, uniqueSecret = normalizeUniqueSecret(uniqueSecret, match.group())
    elif re.search('csp|audit-log', uniqueSecret, re.IGNORECASE):
        ''' matching on known Identity products '''
        productTeam = 'identity'
        pattern = 'csp|audit-log'
        logger.debug('rule match: %s (%s)' % (pattern.upper(), uniqueSecret))
        match = re.search(pattern, uniqueSecret, re.IGNORECASE)
        cmSubService, uniqueSecret = normalizeUniqueSecret(uniqueSecret, match.group())
    elif re.search('policyservice|policy-suggest-service', uniqueSecret, re.IGNORECASE):
        ''' matching on known Diagon Alley products '''
        ##TODO: candidate for release-name-based parsing (e.g. 'infra' is functionally a cm3 value which modifies 'vulnerability-assessment')
        productTeam = normalizeCasing('Diagon Alley')
        pattern = 'policyservice|policy-suggest-service'
        logger.debug('rule match: %s (%s)' % (pattern.upper(), uniqueSecret))
        match = re.search(pattern, uniqueSecret, re.IGNORECASE)
        cmSubService, uniqueSecret = normalizeUniqueSecret(uniqueSecret, match.group())
    elif re.search('cm-cps|CloudfrontKeyPair|cds-cps|cm-rms', uniqueSecret, re.IGNORECASE):
        ''' matching on known Gray Matter products '''
        ##TODO: candidate for release-name-based parsing (e.g. 'infra' is functionally a cm3 value which modifies 'cm-cps')
        productTeam = normalizeCasing('gray-matter')
        pattern = 'cm-cps|CloudfrontKeyPair|cds-cps|cm-rms'
        logger.debug('rule match: %s (%s)' % (pattern.upper(), uniqueSecret))
        match = re.search(pattern, uniqueSecret, re.IGNORECASE)
        cmSubService, uniqueSecret = normalizeUniqueSecret(uniqueSecret, match.group())
    elif re.match('sensor', uniqueSecret, re.IGNORECASE):
        ''' matching on known Dog products '''
        productTeam = 'dog'
        pattern = 'sensor'
        logger.debug('rule match: %s (%s)' % (pattern.upper(), uniqueSecret))
        match = re.search(pattern, uniqueSecret, re.IGNORECASE)
        cmSubService, uniqueSecret = normalizeUniqueSecret(uniqueSecret, match.group())
    ## One-off product matching due to slight pattern nuances
    elif re.match('wild-', uniqueSecret, re.IGNORECASE):
        logger.debug('rule match: %s (%s)' % ('wild-'.upper(), uniqueSecret))
        productTeam = 'prodeng'
        cmSubService, uniqueSecret = normalizeUniqueSecret(uniqueSecret, 'wild')
    elif re.match('cdc-reputation-augmentation', uniqueSecret, re.IGNORECASE):
        logger.debug('rule match: %s (%s)' % ('cdc-reputation-augmentation'.upper(), uniqueSecret))
        productTeam = 'titans'
        cmSubService, uniqueSecret = normalizeUniqueSecret(uniqueSecret, 'augmentation')
    elif re.match('psc-tf-amazon-mq', uniqueSecret, re.IGNORECASE):
        logger.debug('rule match: %s (%s)' % ('psc-tf-amazon-mq'.upper(), uniqueSecret))
        productTeam = 'titans'
        cmSubService = 'eventPipeline'
        uniqueSecret = 'cacheEvictionBroker'
    elif re.match('maxmind|max-mind', uniqueSecret, re.IGNORECASE):
        logger.debug('rule match: %s (%s)' % ('maxmind'.upper(), uniqueSecret))
        elements = parseString(uniqueSecret)
        productTeam = 'maxmind'
        cmSubService = 'geoip'
        remove_me_regex = 'maxmind|max|mind|geoip'
        [elements.remove(item) for item in parseString(uniqueSecret) if re.search(remove_me_regex, item, re.IGNORECASE)]
        uniqueSecret = normalizeCasing(elements) if elements else 'credentials'
    elif re.search('data-forwarder', uniqueSecret, re.IGNORECASE):
        logger.debug('rule match: %s (%s)' % ('data-forwarder'.upper(), uniqueSecret))
        elements = parseString(uniqueSecret)
        ## need to loop 'parseString(uniqueSecret)' vs 'elements' in order to remove more than one match
        exclusions = []
        for item in parseString(uniqueSecret):
            if re.search('kcdev01|kctest03', item, re.IGNORECASE):
                env = item
                exclusions.append(item)
        productTeam = 'cat'
        cmSubService, uniqueSecret = normalizeUniqueSecret(uniqueSecret, 'data-forwarder', exclusions=exclusions)
    ## Default matching based on number of split-able items in the string
    else:
        logger.debug('rule match: %s (%s)' % ('default'.upper(), uniqueSecret))
        elements = parseString(uniqueSecret)
        logger.debug('default elements: %s' % elements)
        if len(elements) == 1:
            productTeam = None
            cmSubService = None
            uniqueSecret = normalizeCasing(uniqueSecret)
        else:
            productTeam = elements[0]
            if len(elements) == 2:
                cmSubService = None
                remove_me_regex = '|'.join(e for e in elements[0:1])
            else:
                cmSubService = elements[1]
                remove_me_regex = '|'.join(e for e in elements[0:2])
            [elements.remove(item) for item in parseString(uniqueSecret) if re.search(remove_me_regex, item, re.IGNORECASE)]
            uniqueSecret = normalizeCasing(elements)
    return env, productTeam, cmSubService, uniqueSecret
